----------------------------------------------------------------------------------
-- Project Name: Gage Hash Function, C: 224 and r: 8
-- 
-- Copyright 2019:
--     Mohamed El-Hadedy <mealy@cpp.edu>
--     Danilo Gligoroski <danilog@ntnu.no>
-- 
-- This program is free software; you can redistribute it and/or
-- modify it under the terms of the GNU General Public License as
-- published by the Free Software Foundation; either version 2 of the
-- License, or (at your option) any later version.
-- 
-- This program is distributed in the hope that it will be useful, but
-- WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
-- General Public License for more details.
-- 
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
-- 02110-1301, USA.
-- 
----------------------------------------------------------------------------------

----------------------------------------------------------------------------------
-- Company: California State Polytechnic University, Pomona, Norwegian University of Science and Technology (NTNU)
-- Engineer-1: Assistant Professor at ECE-department, College of Engineering 
-- Engineer-2: Professor at Department of Information Security and Communication Technology
-- 
-- Create Date: 06/22/2019 02:06:31 PM
-- Design Name-1: Mohamed El-Hadedy, also Airforce Research Lab(AFRL) fellow, US-AFA
-- Design Name-2: Danilo Gligoroski, also Visiting professor: Department of Mathematics and Statistics, University of South Florida, USA

-- Module Name: G_Engine - Behavioral
-- Project Name: Gage Hash Function, C: 224 and r: 8
-- Target Devices: xc7a35tcpg236-1 
-- Tool Versions: Vivado 2019.1
-- Description: Hash function Engine (Hash Enge processor)
-- Dependencies: 
-- 
-- Revision:
-- Revision 0.01 - File Created
-- Additional Comments:
-- 
----------------------------------------------------------------------------------


library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.numeric_std.all;
use ieee.STD_LOGIC_ARITH.all;
use ieee.std_logic_unsigned.all;
use IEEE.math_real.all;
-- Uncomment the following library declaration if using
-- arithmetic functions with Signed or Unsigned values

-- Uncomment the following library declaration if instantiating
-- any Xilinx leaf cells in this code.
--library UNISIM;
--use UNISIM.VComponents.all;

entity Dtransform is
    generic (C: positive:= 224; r: positive:=8); 
    Port ( clk      : in  STD_LOGIC;
           rst      : in  STD_LOGIC;
           rst_auto : in  STD_LOGIC;
           STRT     : in  STD_LOGIC; 
           OS       : in  STD_LOGIC_VECTOR (c+r-1 downto 0);
           NS       : out STD_LOGIC_VECTOR (c+r-1 downto 0);
           EN_NS_OUT: out STD_LOGIC
         );
end Dtransform;

architecture Behavioral of Dtransform is
constant b: std_logic_vector(7 downto 0):= "11101000";


component Leader
    Port ( LDR_INP : in  STD_LOGIC_VECTOR (4 downto 0);
           LDR_OP  : out STD_LOGIC_VECTOR (1 downto 0)
         );
end component;

component Q_sbox 
  Port   (
         clk   : in  std_logic;
         rst   : in  std_logic; 
         EN_IN : in  std_logic; 
         INP   : in  std_logic_vector(3 downto 0);
         OP    : out std_logic_vector(1 downto 0)
         );
end component;


signal X     : std_logic_vector(c+r-1 downto 0); 
signal Y     : std_logic_vector(c+r-1 downto 0);
--signal tmp   : std_logic_vector(3 downto 0);
--signal tmp_op: std_logic_vector(1 downto 0); 
signal EN_OUT_auto: std_logic; 
signal EN_IN : std_logic;
signal OS_VAR: std_logic_vector(463 downto 0); 
signal NS_VAR: std_logic_Vector(231 downto 0); 
signal L     : std_logic_vector(1 downto 0);
signal P     : std_logic_vector(c+r-1 downto 0); 
--signal LSEL  : std_logic;
signal BSHFL : std_logic; 
signal LDR_PTR: std_logic_vector(4 downto 0); 
signal round_count: std_logic_vector(4 downto 0); 

signal OUT_WIRE: std_logic; 
signal EN_NS_OUT_TMP:std_logic;

begin


LDR: leader port map (
                        LDR_INP => LDR_PTR,
                        LDR_OP  => L
                     ); 

OS_VAR(0) <= X(1);
OS_VAR(1) <= X(0);


OS_VAR(3) <= (L(1)); 
OS_VAR(2) <= (L(0)); 


MUX_NETWORK: for i_mux in 0 to (conv_integer(c+r-1)) generate

            X(i_mux) <= (BSHFL and P(i_mux)) or ( not BSHFL and OS(i_mux));
 
end generate; 


DELAY:process(clk,rst)
      begin 
        if (rst = '1') then 
         EN_OUT_auto <='0';
        elsif (rising_edge(clk)) then 
         EN_OUT_auto <= EN_IN; 
        end if; 
      end process; 

Q0: Q_sbox port map (
                     clk    => clk,
                     rst    => rst, 
                     EN_IN  => EN_IN,
                     INP    => OS_VAR(3 downto 0),
                     OP     => NS_VAR(1 downto 0)
                    );


      Y(0) <= NS_VAR(1);
      Y(1) <= NS_VAR(0);  

      
ROUND: for ii in 0 to (conv_integer('0'& b(7 downto 1))-2) generate
        OS_VAR(3+4*(ii+1)) <= X( conv_integer((0 +(ii*2))));   --OS_VAR(3+4*(ii+1))
        OS_VAR(2+4*(ii+1)) <= X( conv_integer((1 +(ii*2)))); --OS_VAR(2+4*(ii+1))
        OS_VAR(1+4*(ii+1)) <= X( conv_integer((2 +(ii*2)))); --OS_VAR(1+4*(ii+1))
        OS_VAR(0+4*(ii+1)) <= X( conv_integer((3 +(ii*2)))); --OS_VAR(0+4*(ii+1))
        
         QX: Q_sbox port map (
                              clk  => clk,
                              rst  => rst, 
                              EN_IN=> EN_IN,  
                              INP  => OS_VAR((3+4*(ii+1)) downto (0+4*(ii+1))),
                              OP   => NS_VAR((1+(ii+1)*2) downto (ii+1)*2) 
                            );
        Y((2*(ii+1)+0)) <=  NS_VAR(1+(ii+1)*2); 
        Y((2*(ii+1)+1)) <=  NS_VAR(0+(ii+1)*2);                     
end generate; 

                 p( 0  )<= Y(56);
                 p( 1  )<= Y(42);
                 p( 2  )<= Y(28);
                 p( 3  )<= Y(7);
                 p( 4  )<= Y(14);
                 p( 5  )<= Y(21);
                 p( 6  )<= Y(49);
                 p( 7  )<= Y(35);
                 p( 8  )<= Y(64);
                 p( 9  )<= Y(50);
                 p( 10 )<= Y(36);
                 p( 11 )<= Y(15);
                 p( 12 )<= Y(22);
                 p( 13 )<= Y(29);
                 p( 14 )<= Y(57);
                 p( 15 )<= Y(43);
                 p( 16 )<= Y(72);
                 p( 17 )<= Y(58);
                 p( 18 )<= Y(44);
                 p( 19 )<= Y(23);
                 p( 20 )<= Y(30);
                 p( 21 )<= Y(37);
                 p( 22 )<= Y(65);
                 p( 23 )<= Y(51);
                 p( 24 )<= Y(80);
                 p( 25 )<= Y(66);
                 p( 26 )<= Y(52);
                 p( 27 )<= Y(31);
                 p( 28 )<= Y(38);
                 p( 29 )<= Y(45);
                 p( 30 )<= Y(73);
                 p( 31 )<= Y(59);
                 p( 32 )<= Y(88);
                 p( 33 )<= Y(74);
                 p( 34 )<= Y(60);
                 p( 35 )<= Y(39);
                 p( 36 )<= Y(46);
                 p( 37 )<= Y(53);
                 p( 38 )<= Y(81);
                 p( 39 )<= Y(67);
                 p( 40 )<= Y(96);
                 p( 41 )<= Y(82);
                 p( 42 )<= Y(68);
                 p( 43 )<= Y(47);
                 p( 44 )<= Y(54);
                 p( 45 )<= Y(61);
                 p( 46 )<= Y(89);
                 p( 47 )<= Y(75);
                 p( 48 )<= Y(104);
                 p( 49 )<= Y( 90);
                 p( 50 )<= Y( 76);
                 p( 51 )<= Y(  55);
                 p( 52 )<= Y(  62);
                 p( 53 )<= Y(  69);
                 p( 54 )<= Y(  97);
                 p( 55 )<= Y(  83);
                 p( 56 )<= Y( 112);
                 p( 57 )<= Y(  98);
                 p( 58 )<= Y(  84);
                 p( 59 )<= Y(  63);
                 p( 60 )<= Y(  70);
                 p( 61 )<= Y(  77);
                 p( 62 )<= Y(  105);
                 p( 63 )<= Y(  91);
                 p( 64 )<= Y( 120);
                 p( 65 )<= Y(  106);
                 p( 66 )<= Y(  92);
                 p( 67 )<= Y(  71);
                 p( 68 )<= Y(  78);
                 p( 69 )<= Y(  85);
                 p( 70 )<= Y(  113);
                 p( 71 )<= Y(  99);
                 p( 72 )<= Y(  128);
                 p( 73 )<= Y(  114);
                 p( 74 )<= Y(  100);
                 p( 75 )<= Y(  79);
                 p( 76 )<= Y(  86);
                 p( 77 )<= Y(  93);
                 p( 78 )<= Y(  121);
                 p( 79 )<= Y(  107);
                 p( 80 )<= Y(  136);
                 p( 81 )<= Y(  122);
                 p( 82 )<= Y(  108);
                 p( 83 )<= Y(  87);
                 p( 84 )<= Y(  94);
                 p( 85 )<= Y(  101);
                 p( 86 )<= Y(  129);
                 p( 87 )<= Y(  115);
                 p( 88 )<= Y(  144);
                 p( 89 )<= Y(  130);
                 p( 90 )<= Y(  116);
                 p( 91 )<= Y(  95);
                 p( 92 )<= Y(  102);
                 p( 93 )<= Y(  109);
                 p( 94 )<= Y(  137);
                 p( 95 )<= Y(  123);
                 p( 96 )<= Y(  152);
                 p( 97 )<= Y(  138);
                 p( 98 )  <= Y(124);
                 p( 99 )  <= Y(103);
                 p( 100 )<= Y(110);
                 p( 101 )<= Y(117);
                 p( 102 )<= Y(145);
                 p( 103 )<= Y(131);
                 p( 104 )<= Y(160);
                 p( 105 )<= Y(146);
                 p( 106 )<= Y(132);
                 p( 107 )<= Y(111);
                 p( 108 )<= Y(118);
                 p( 109 )<= Y(125);
                 p( 110 )<= Y(153);
                 p( 111 )<= Y(139);
                 p( 112 )<= Y(168);
                 p( 113 )<= Y(154);
                 p( 114 )<= Y(140);
                 p( 115 )<= Y(119);
                 p( 116 )<= Y(126);
                 p( 117 )<= Y(133);
                 p( 118 )<= Y(161);
                 p( 119 )<= Y(147);
                 p( 120 )<= Y(176);
                 p( 121 )<= Y(162);
                 p( 122 )<= Y(148);
                 p( 123 )<= Y(127);
                 p( 124 )<= Y(134);
                 p( 125 )<= Y(141);
                 p( 126 )<= Y(169);
                 p( 127 )<= Y(155);
                 p( 128 )<= Y(184);
                 p( 129 )<= Y(170);
                 p( 130 )<= Y(156);
                 p( 131 )<= Y(135);
                 p( 132 )<= Y(142);
                 p( 133 )<= Y(149);
                 p( 134 )<= Y(177);
                 p( 135 )<= Y(163);
                 p( 136 )<= Y(192);
                 p( 137 )<= Y(178);
                 p( 138 )<= Y(164);
                 p( 139 )<= Y(143);
                 p( 140 )<= Y(150);
                 p( 141 )<= Y(157);
                 p( 142 )<= Y(185);
                 p( 143 )<= Y(171);
                 p( 144 )<= Y(200);
                 p( 145 )<= Y(186);
                 p( 146 )<= Y(172);
                 p( 147 )<= Y(151);
                 p( 148 )<= Y( 158);
                 p( 149 )<= Y( 165);
                 p( 150 )<= Y( 193);
                 p( 151 )<= Y( 179);
                 p( 152 )<= Y(208);
                 p( 153 )<= Y( 194);
                 p( 154 )<= Y( 180);
                 p( 155 )<= Y( 159);
                 p( 156 )<= Y( 166);
                 p( 157 )<= Y( 173);
                 p( 158 )<= Y( 201);
                 p( 159 )<= Y( 187);
                 p( 160 )<= Y( 216);
                 p( 161 )<= Y( 202);
                 p( 162 )<= Y( 188);
                 p( 163 )<= Y( 167);
                 p( 164 )<= Y( 174);
                 p( 165 )<= Y( 181);
                 p( 166 )<= Y( 209);
                 p( 167 )<= Y( 195);
                 p( 168 )<= Y( 224);
                 p( 169 )<= Y( 210);
                 p( 170 )<= Y( 196);
                 p( 171 )<= Y( 175);
                 p( 172 )<= Y( 182);
                 p( 173 )<= Y( 189);
                 p( 174 )<= Y( 217);
                 p( 175 )<= Y( 203);
                 p( 176 )<= Y( 0);
                 p( 177 )<= Y( 218);
                 p( 178 )<= Y( 204);
                 p( 179 )<= Y( 183);
                 p( 180 )<= Y( 190);
                 p( 181 )<= Y( 197);
                 p( 182 )<= Y( 225);
                 p( 183 )<= Y( 211);
                 p( 184 )<= Y( 8);
                 p( 185 )<= Y( 226);
                 p( 186 )<= Y( 212);
                 p( 187 )<= Y( 191);
                 p( 188 )<= Y( 198);
                 p( 189 )<= Y( 205);
                 p( 190 )<= Y( 1);
                 p( 191 )<= Y( 219);
                 p( 192 )<= Y( 16);
                 p( 193 )<= Y( 2);
                 p( 194 )<= Y( 220);
                 p( 195 )<= Y( 199);
                 p( 196 )<= Y( 206);
                 p( 197 )<= Y( 213);
                 p( 198 )<= Y( 9);
                 p( 199 )<= Y( 227);
                 p( 200 )<= Y( 24);
                 p( 201 )<= Y( 10);
                 p( 202 )<= Y( 228);
                 p( 203 )<= Y( 207);
                 p( 204 )<= Y( 214);
                 p( 205 )<= Y( 221);
                 p( 206 )<= Y( 17);
                 p( 207 )<= Y( 3);
                 p( 208 )<= Y( 32);
                 p( 209 )<= Y( 18);
                 p( 210 )<= Y( 4);
                 p( 211 )<= Y( 215);
                 p( 212 )<= Y( 222);
                 p( 213 )<= Y( 229);
                 p( 214 )<= Y( 25);
                 p( 215 )<= Y( 11);
                 p( 216 )<= Y( 40);
                 p( 217 )<= Y( 26);
                 p( 218 )<= Y( 12);
                 p( 219 )<= Y( 223);
                 p( 220 )<= Y( 230);
                 p( 221 )<= Y( 5);
                 p( 222 )<= Y( 33);
                 p( 223 )<= Y( 19);
                 p( 224 )<= Y( 48);
                 p( 225 )<= Y( 34);
                 p( 226 )<= Y( 20);
                 p( 227 )<= Y( 231);
                 p( 228 )<= Y( 6);
                 p( 229 )<= Y( 13);
                 p( 230 )<= Y( 41);
                 p( 231 )<= Y( 27);   
LDR_PTR <= round_count;     

BSHFL <=   (not STRT and  (round_count(0) or round_count(1) or round_count(2) or round_count(3) or round_count(4)));     
                 
FSM_PROCESS:process(clk,rst)
            begin 
                if(rst = '1') then 
                    round_count <=(others =>'0');
                    EN_IN       <= '0';
                elsif(rising_edge(clk)) then 
                  if (STRT = '1'  ) then
                       EN_IN      <= '1';
                       round_count <= (others =>'0');
                  elsif (EN_OUT_auto = '1') then 
                       round_count <= round_count +1;
                       
                  end if;
                  if (EN_NS_OUT_TMP =  '1') then
                    EN_IN <= '0';
                    round_count <= (others =>'0');
                  end if;  
                end if; 
            end process;   
            
DELAY_CONTROL:process(clk,rst)
              begin 
                if(rst = '1') then 
                    OUT_WIRE <= '0';
                elsif(rising_edge(clk)) then 
                    OUT_WIRE <= round_count(0) and round_count(1) and round_count(2) and round_count(3) and round_count(4);
                end if; 
              end process;                     
            
FSM_OUTPUT:process(clk,rst)
           begin 
            if (rst = '1') then 
                EN_NS_OUT_TMP <= '0';
            elsif (rising_edge(clk)) then 
                if (OUT_WIRE = '1') then
                   EN_NS_OUT_TMP <= '1';
                else 
                   EN_NS_OUT_TMP <= '0';
                end if; 
            end if; 
           end process;   
           
FSM_OUT_REG: process(clk,rst)
             begin 
             if (rst_auto = '1') then 
                NS <= (others =>'0');
            elsif (rising_edge(clk)) then 
                if (OUT_WIRE = '1') then
                   NS        <=   Y; 
                end if; 
            end if;             
             
             
             end process;                             
EN_NS_OUT <= EN_NS_OUT_TMP;
end Behavioral;
