----------------------------------------------------------------------------------
-- Project Name: Gage Hash Function, C: 224 and r: 8
-- 
-- Copyright 2019:
--     Mohamed El-Hadedy <mealy@cpp.edu>
--     Danilo Gligoroski <danilog@ntnu.no>
-- 
-- This program is free software; you can redistribute it and/or
-- modify it under the terms of the GNU General Public License as
-- published by the Free Software Foundation; either version 2 of the
-- License, or (at your option) any later version.
-- 
-- This program is distributed in the hope that it will be useful, but
-- WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
-- General Public License for more details.
-- 
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
-- 02110-1301, USA.
-- 
----------------------------------------------------------------------------------

----------------------------------------------------------------------------------
-- Company: California State Polytechnic University, Pomona, Norwegian University of Science and Technology (NTNU)
-- Engineer-1: Assistant Professor at ECE-department, College of Engineering 
-- Engineer-2: Professor at Department of Information Security and Communication Technology
-- 
-- Create Date: 06/22/2019 02:06:31 PM
-- Design Name-1: Mohamed El-Hadedy, also Airforce Research Lab(AFRL) fellow, US-AFA
-- Design Name-2: Danilo Gligoroski, also Visiting professor: Department of Mathematics and Statistics, University of South Florida, USA

-- Module Name: G_Engine - Behavioral
-- Project Name: Gage Hash Function , C: 224 and r: 8
-- Target Devices: xc7a35tcpg236-1 
-- Tool Versions: Vivado 2019.1
-- Description: Hash function Engine (SBOX 4-TO-2)
-- 
-- Dependencies: 
-- 
-- Revision:
-- Revision 0.01 - File Created
-- Additional Comments:
-- 
----------------------------------------------------------------------------------


library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use ieee.STD_LOGIC_ARITH.all;
use ieee.std_logic_unsigned.all;
use IEEE.math_real.all;
-- Uncomment the following library declaration if using
-- arithmetic functions with Signed or Unsigned values
--use IEEE.NUMERIC_STD.ALL;

-- Uncomment the following library declaration if instantiating
-- any Xilinx leaf cells in this code.
--library UNISIM;
--use UNISIM.VComponents.all;

entity Q_sbox is
  Port   (
         clk   : in  std_logic;
         rst   : in  std_logic; 
         EN_IN : in  std_logic; 
         INP   : in  std_logic_vector(3 downto 0);
         OP    : out std_logic_vector(1 downto 0)
         );
end Q_sbox;

architecture Behavioral of Q_sbox is



begin



-- here we need a LUT to setup the functionality of the Q-sbox 

SBOX: process(clk,rst) begin
    if (rst = '1') then 
        OP <= ("00");
    elsif (rising_edge(clk)) then 
    if (EN_IN = '1') then 
    cont:case (INP) is 
        when "0000" => OP <= "01";
        when "0001" => OP <= "00";
        when "0010" => OP <= "11";
        when "0011" => OP <= "10";
        when "0100" => OP <= "00";
        when "0101" => OP <= "10";
        when "0110" => OP <= "01";
        when "0111" => OP <= "11";
        when "1000" => OP <= "10";
        when "1001" => OP <= "11";
        when "1010" => OP <= "00";
        when "1011" => OP <= "01";
        when "1100" => OP <= "11";
        when "1101" => OP <= "01";
        when "1110" => OP <= "10";
        when "1111" => OP <= "00";
        when others => OP <= (others =>'Z');
    end case cont; 
    end if; 
   end if; 
end process SBOX; 





end Behavioral;
