----------------------------------------------------------------------------------
-- Project Name: Gage Hash Function, C: 224 and r: 8
-- 
-- Copyright 2019:
--     Mohamed El-Hadedy <mealy@cpp.edu>
--     Danilo Gligoroski <danilog@ntnu.no>
-- 
-- This program is free software; you can redistribute it and/or
-- modify it under the terms of the GNU General Public License as
-- published by the Free Software Foundation; either version 2 of the
-- License, or (at your option) any later version.
-- 
-- This program is distributed in the hope that it will be useful, but
-- WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
-- General Public License for more details.
-- 
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
-- 02110-1301, USA.
-- 
----------------------------------------------------------------------------------

----------------------------------------------------------------------------------
-- Company: California State Polytechnic University, Pomona, Norwegian University of Science and Technology (NTNU)
-- Engineer-1: Assistant Professor at ECE-department, College of Engineering 
-- Engineer-2: Professor at Department of Information Security and Communication Technology
-- 
-- Create Date: 06/22/2019 02:06:31 PM
-- Design Name-1: Mohamed El-Hadedy, also Airforce Research Lab(AFRL) fellow, US-AFA
-- Design Name-2: Danilo Gligoroski, also Visiting professor: Department of Mathematics and Statistics, University of South Florida, USA

-- Module Name: G_Engine - Behavioral
-- Project Name: Gage Hash Function , C: 224 and r: 8
-- Target Devices: xc7a35tcpg236-1 
-- Tool Versions: Vivado 2019.1
-- Description: Hash function Engine
-- 
-- Dependencies: FIFO (STD_FIFO.vhd), Hash_Engine (Dtransform.vhd), LDR (leader.vhd), Round (Q_sbox.vhd)
-- 
-- Revision:
-- Revision 0.01 - File Created
-- Additional Comments:
-- 
----------------------------------------------------------------------------------


library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.numeric_std.all;
use ieee.STD_LOGIC_ARITH.all;
use ieee.std_logic_unsigned.all;
use IEEE.math_real.all;

-- Uncomment the following library declaration if using
-- arithmetic functions with Signed or Unsigned values
--use IEEE.NUMERIC_STD.ALL;

-- Uncomment the following library declaration if instantiating
-- any Xilinx leaf cells in this code.
--library UNISIM;
--use UNISIM.VComponents.all;

entity gage1h256c224r008AllParallel is
    generic (C: positive:= 224; r: positive:=8); 
    Port ( clk     : in  STD_LOGIC;
           rst     : in  STD_LOGIC;
           M_ST    : in  STD_LOGIC; 
           wr_en   : in  STD_LOGIC;
           Message : in  STD_LOGIC_VECTOR (7 downto 0);
           HASH_MSG: out STD_LOGIC_VECTOR (7 downto 0);
           HASH_OUT: out STD_LOGIC
         );
end gage1h256c224r008AllParallel;

architecture Behavioral of gage1h256c224r008AllParallel is
component Dtransform 
    generic (C: positive:= 224; r: positive:=64); 
    Port ( clk      : in  STD_LOGIC;
           rst      : in  STD_LOGIC;
           rst_auto : in  STD_LOGIC;
           STRT     : in  STD_LOGIC; 
           OS       : in  STD_LOGIC_VECTOR (c+r-1 downto 0);
           NS       : out STD_LOGIC_VECTOR (c+r-1 downto 0);
           EN_NS_OUT: out STD_LOGIC
         );
end component;

component fifo_generator_0 
    PORT (
           clk   : IN  STD_LOGIC;
           rst   : IN  STD_LOGIC;
           din   : IN  STD_LOGIC_VECTOR(7 DOWNTO 0);
           wr_en : IN  STD_LOGIC;
           rd_en : IN  STD_LOGIC;
           dout  : OUT STD_LOGIC_VECTOR(7 DOWNTO 0);
           full  : OUT STD_LOGIC;
           empty : OUT STD_LOGIC
          );
   END component;
   
component STD_FIFO 
	Generic (
		constant DATA_WIDTH  : positive := 8;
		constant FIFO_DEPTH	: positive := 8
	);
	Port ( 
		CLK		: in  STD_LOGIC;
		RST		: in  STD_LOGIC;
		WriteEn	: in  STD_LOGIC;
		DataIn	: in  STD_LOGIC_VECTOR (DATA_WIDTH - 1 downto 0);
		ReadEn	: in  STD_LOGIC;
		DataOut	: out STD_LOGIC_VECTOR (DATA_WIDTH - 1 downto 0);
		Empty	: out STD_LOGIC;
		Full	: out STD_LOGIC
	);
end component;   

signal rd_en: std_logic; 
signal fifo_out: std_logic_vector(7 downto 0); 
signal empty, full: std_logic; 

signal ENG_STRT : std_logic; 
signal OS_ENGINE: std_logic_vector(c+r-1 downto 0);
signal OS_ENGINE_tmp: std_logic_vector(r-1 downto 0); 

signal NS       : std_logic_vector(c+r-1 downto 0);
signal EN_NS_OUT: std_logic; 
constant FINL   : std_logic_vector(7 downto 0):= "10000000";
signal X        : std_logic_vector(7 downto 0); 

signal FIN_SEL  : std_logic;
signal HASH_SEL : std_logic; 

signal count_fsm: std_logic_vector(5 downto 0); 

signal count_out: std_logic_vector(5 downto 0); 
signal rst_auto : std_logic;

signal rst_eng_auto: std_logic; 
signal rst_mux:std_logic; 

TYPE State_type IS (INIT, MSG, EMPY, DONE); 

signal nxt_fsm, ps_fsm:state_type;

signal HASH_OUT_FINAL, HASH_OUT_FINAL_TMP: std_logic; 

signal  HASH_MSG_TMP: std_logic_vector(r-1 downto 0);
begin

rst_mux <= rst or rst_eng_auto;

FIFO_ENG: STD_FIFO generic map (DATA_WIDTH => 8, FIFO_DEPTH => 64)
                   	Port   map ( 
                                CLK		=> clk,
                                RST		=> rst,
                                WriteEn	=> wr_en,
                                DataIn	=> Message,
                                ReadEn	=> rd_en,
                                DataOut	=> fifo_out,
                                Empty	=> empty,
                                Full	=> full 
                            
                              );       

--FIFO: fifo_generator_0 port map (
--                                   clk   => clk,
--                                   rst   => rst,
--                                   din   => Message, 
--                                   wr_en => wr_en,
--                                   rd_en => rd_en,
--                                   dout  => fifo_out,
--                                   full  => full,
--                                   empty => empty
--                                );

HASH_ENGINE: Dtransform generic map (c=> c, r=>r)
                        port    map (
                                     clk       => clk,
                                     rst       => rst_mux,
                                     rst_auto  => rst_auto,
                                     STRT      => ENG_STRT,
                                     OS        => OS_ENGINE,
                                     NS        => NS,
                                     EN_NS_OUT => EN_NS_OUT
                                    );
MUX_NETWORK: for i_mux in 0 to (conv_integer(r-1)) generate
            X(i_mux) <= (FIN_SEL and FINL(i_mux)) or ( not FIN_SEL and fifo_out(i_mux));
end generate;
MUX_MSG: for i in 0 to (conv_integer(r-1)) generate
            OS_ENGINE_tmp(i) <= NS(i) xor X(7-i) ;
end generate;


MUX_STAT_MIX: for j_mux in 0 to (conv_integer(r-1)) generate
              OS_ENGINE(j_mux) <= (HASH_SEL and NS(j_mux)) or ( not HASH_SEL and OS_ENGINE_tmp(j_mux));
end generate;




ROUND: for ii in 8 to (conv_integer(c+r-1)) generate
         OS_ENGINE(ii)<= NS(ii);
end generate; 



FSM_MOV: process(clk,rst)
         begin 
            if (rst = '1') then 
                ps_fsm <= INIT;
            elsif(rising_edge(clk)) then 
                ps_fsm <= nxt_fsm; 
            end if; 
         
         end process; 
         
MUX_STAT_ORD: for w in 0 to (conv_integer(r-1)) generate
              HASH_MSG_TMP(7-w) <= NS(w); 
end generate; 

FSM_HASH_OP: process(clk,rst)
             begin 
             
             if(rst = '1') then 
               count_fsm <= (others =>'0');
               rst_auto  <= '1';
               rst_eng_auto <= '0';
               rd_en     <= '0';
               FIN_SEL   <= '0';
               HASH_SEL  <= '0';
               HASH_OUT  <= '0';
               HASH_MSG  <= (others =>'0');
               count_out <= (others =>'0');

             elsif(rising_edge(clk)) then 
                case ps_fsm is
                    when INIT  =>
                        rst_eng_auto <= '1';
                        rst_auto     <= '1';
                        ENG_STRT     <= '0';
                        FIN_SEL      <= '0';
                        HASH_SEL     <= '0';  
                        rd_en        <= '0';
                        count_out    <= (others =>'0');
                        count_fsm    <= (others=>'0');
                       if (M_ST = '1') then 
                        rst_auto     <='0';
                           if (empty = '1') then 
                            nxt_fsm <= EMPY;
                            count_fsm <= "000010";
                           else 
                            nxt_fsm <= MSG;
                            count_fsm <= (others =>'0');
                           end if; 
                       else
                           nxt_fsm <= INIT; 
                       end if; 
                       HASH_OUT <= EN_NS_OUT;

                    
                    when MSG   =>
                        if (count_fsm = 0) then 
                            rst_eng_auto <= '0';
                            ENG_STRT <= '0';
                            rd_en    <= '1';
                            FIN_SEL  <= '0';
                            HASH_SEL <= '0';                             
                            count_fsm <= "000001";                            
                        elsif(count_fsm = 1) then
                            ENG_STRT <= '1';
                            rd_en    <= '0';
                            count_fsm <= "000010";
                        elsif(count_fsm = 2) then      
                            ENG_STRT <= '0';
                          if (EN_NS_OUT = '1') then    
                            if (empty = '1') then 
                              nxt_fsm    <= EMPY; 
                              rst_eng_auto <= '1';
                            else 
                              nxt_fsm    <= MSG;
                              count_fsm  <= "000000";
                            end if; 
                          end if;
                       end if;   
                       
                    when EMPY  =>
                        if(count_fsm = 2) then 
                            rst_eng_auto <= '0';
                            ENG_STRT <= '0';
                            rd_en    <= '0';
                            FIN_SEL  <= '1';
                            HASH_SEL <= '0';                             
                            count_fsm <= count_fsm - 1;                            
                        elsif (count_fsm = 1) then 
                            ENG_STRT <= '1';
                            count_fsm <= count_fsm - 1;
                        elsif (count_fsm = 0) then    
                           ENG_STRT <= '0';
                          if (EN_NS_OUT = '1') then       
                            nxt_fsm    <= DONE; 
                            HASH_MSG <= HASH_MSG_TMP;
                            rst_eng_auto <= '1';
                          else 
                            nxt_fsm    <= EMPY;
                            count_fsm  <= (others =>'0');
                          end if; 
                           HASH_OUT <= EN_NS_OUT;  
                        end if;                             
                    when  DONE  =>  
                        if (count_fsm = 0) then 
                            rst_eng_auto <= '0';
                            ENG_STRT <= '0';
                            rd_en    <= '0';
                            FIN_SEL  <= '0';
                            HASH_SEL <= '1';                             
                            count_fsm <= count_fsm + 1;                            
                        elsif(count_fsm = 1) then
                            ENG_STRT <= '1';
                            count_fsm <= count_fsm + 1;
                        elsif(count_fsm = 2) then      
                            ENG_STRT <= '0';
                          if (EN_NS_OUT = '1') then    
                            HASH_MSG <= HASH_MSG_TMP; 
                            count_out <= count_out +1;
                            rst_eng_auto <= '1';
                            count_fsm <= (others =>'0');
                          end if;
                       end if;   
                         HASH_OUT <= EN_NS_OUT;  

                       if (HASH_OUT_FINAL = '1') then 
                          nxt_fsm <= INIT; 
                       else
                        nxt_fsm <= DONE; 
                       end if; 

                when others => 
                end case; 
             
             end if; 
             
             end process; 
             
DELAY_D0: process(clk,rst)
       begin 
        if(rst = '1') then 
            HASH_OUT_FINAL <= '0';
        elsif(rising_edge(clk)) then 
            HASH_OUT_FINAL <=  ( count_out(0) and  count_out(1) and  count_out(2) and  count_out(3) and  count_out(4));  
        end if; 
       
       end process;             

DELAY_D1: process(clk,rst)
       begin 
        if(rst = '1') then 
            HASH_OUT_FINAL_TMP <= '0';
        elsif(rising_edge(clk)) then 
            HASH_OUT_FINAL_TMP <= HASH_OUT_FINAL;  
        end if; 
       
       end process;             
             
             




end Behavioral;
