----------------------------------------------------------------------------------
-- Project Name: Gage Hash Function, C: 224 and r: 8
-- 
-- Copyright 2019:
--     Mohamed El-Hadedy <mealy@cpp.edu>
--     Danilo Gligoroski <danilog@ntnu.no>
-- 
-- This program is free software; you can redistribute it and/or
-- modify it under the terms of the GNU General Public License as
-- published by the Free Software Foundation; either version 2 of the
-- License, or (at your option) any later version.
-- 
-- This program is distributed in the hope that it will be useful, but
-- WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
-- General Public License for more details.
-- 
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
-- 02110-1301, USA.
-- 
----------------------------------------------------------------------------------

----------------------------------------------------------------------------------
-- Company: California State Polytechnic University, Pomona, Norwegian University of Science and Technology (NTNU)
-- Engineer-1: Assistant Professor at ECE-department, College of Engineering 
-- Engineer-2: Professor at Department of Information Security and Communication Technology
-- 
-- Create Date: 06/22/2019 02:06:31 PM
-- Design Name-1: Mohamed El-Hadedy, also Airforce Research Lab(AFRL) fellow, US-AFA
-- Design Name-2: Danilo Gligoroski, also Visiting professor: Department of Mathematics and Statistics, University of South Florida, USA

-- Module Name: G_Engine - Behavioral
-- Project Name: Gage Hash Function , C: 224 and r: 8
-- Target Devices: xc7a35tcpg236-1 
-- Tool Versions: Vivado 2019.1
-- Description: Testing the engine without involving the INitialization, Absorbing, Squeezing operations (CHeck page number 10 , Gage and In Gage submission)
-- 
-- Dependencies: 
-- 
-- Revision:
-- Revision 0.01 - File Created
-- Additional Comments:
-- 
----------------------------------------------------------------------------------


library IEEE;
use IEEE.STD_LOGIC_1164.ALL;

-- Uncomment the following library declaration if using
-- arithmetic functions with Signed or Unsigned values
--use IEEE.NUMERIC_STD.ALL;

-- Uncomment the following library declaration if instantiating
-- any Xilinx leaf cells in this code.
--library UNISIM;
--use UNISIM.VComponents.all;

entity ENGINE_TB is
--  Port ( );
end ENGINE_TB;

architecture Behavioral of ENGINE_TB is
component Dtransform 
    generic (C: positive:= 224; r: positive:=8); 
    Port ( clk      : in  STD_LOGIC;
           rst      : in  STD_LOGIC;
           rst_auto : in  STD_LOGIC;
           STRT     : in  STD_LOGIC; 
           OS       : in  STD_LOGIC_VECTOR (c+r-1 downto 0);
           NS       : out STD_LOGIC_VECTOR (c+r-1 downto 0);
           EN_NS_OUT: out STD_LOGIC
         );
end component;

constant clk_period: time:= 10 ns; 

signal clk : std_logic; 
signal rst : std_logic;
signal STRT: std_logic; 

signal OS  : std_logic_vector(231 downto 0); 
signal NS  : std_logic_vector(231 downto 0); 
signal rst_auto:std_logic;
signal EN_NS_OUT:std_logic; 

begin
ENG: Dtransform generic map (c=> 224, r => 8)
                port    map (
                             clk       => clk,
                             rst       => rst, 
                             rst_auto  => rst_auto,
                             STRT      => STRT,
                             OS        => OS,
                             NS        => NS,
                             EN_NS_OUT => EN_NS_OUT
                            );
                            
CLK_GEN: process
         begin 
            clk <= '1';
            wait for clk_period/2;
            clk <= '0';
            wait for clk_period/2;
         end process; 
         
ENG_OPRTN: process
           begin 
            rst  <= '1';
            rst_auto <= '1';
            STRT <= '0';
            OS   <= (others =>'0');
            wait for clk_period;
            rst_auto <= '0'; 
            rst  <= '0';
            STRT <= '1';
            wait for clk_period; 
            STRT <= '0';
            wait for 34* clk_period; 
            rst  <= '1';
            wait for clk_period;
            rst <= '0';
            wait; 
           end process; 

end Behavioral;
