----------------------------------------------------------------------------------
-- Project Name: Gage Hash Function, C: 224 and r: 8
-- 
-- Copyright 2019:
--     Mohamed El-Hadedy <mealy@cpp.edu>
--     Danilo Gligoroski <danilog@ntnu.no>
-- 
-- This program is free software; you can redistribute it and/or
-- modify it under the terms of the GNU General Public License as
-- published by the Free Software Foundation; either version 2 of the
-- License, or (at your option) any later version.
-- 
-- This program is distributed in the hope that it will be useful, but
-- WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
-- General Public License for more details.
-- 
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
-- 02110-1301, USA.
-- 
----------------------------------------------------------------------------------

----------------------------------------------------------------------------------
-- Company: California State Polytechnic University, Pomona, Norwegian University of Science and Technology (NTNU)
-- Engineer-1: Assistant Professor at ECE-department, College of Engineering 
-- Engineer-2: Professor at Department of Information Security and Communication Technology
-- 
-- Create Date: 06/22/2019 02:06:31 PM
-- Design Name-1: Mohamed El-Hadedy, also Airforce Research Lab(AFRL) fellow, US-AFA
-- Design Name-2: Danilo Gligoroski, also Visiting professor: Department of Mathematics and Statistics, University of South Florida, USA

-- Module Name: G_Engine - Behavioral
-- Project Name: Gage Hash Function , C: 224 and r: 8
-- Target Devices: xc7a35tcpg236-1 
-- Tool Versions: Vivado 2019.1
-- Description: Testing the fifo buffer operations 
-- 
-- Dependencies: 
-- 
-- Revision:
-- Revision 0.01 - File Created
-- Additional Comments:
-- 
----------------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use STD.textio.all;
use ieee.std_logic_textio.all;
-- Uncomment the following library declaration if using
-- arithmetic functions with Signed or Unsigned values
--use IEEE.NUMERIC_STD.ALL;

-- Uncomment the following library declaration if instantiating
-- any Xilinx leaf cells in this code.
--library UNISIM;
--use UNISIM.VComponents.all;

entity FIFO_TB is
--  Port ( );
end FIFO_TB;

architecture Behavioral of FIFO_TB is

      component fifo_generator_0 
      PORT (
            clk   : IN STD_LOGIC;
            rst   : IN STD_LOGIC;
            din   : IN STD_LOGIC_VECTOR(7 DOWNTO 0);
            wr_en : IN STD_LOGIC;
            rd_en : IN STD_LOGIC;
            dout  : OUT STD_LOGIC_VECTOR(7 DOWNTO 0);
            full  : OUT STD_LOGIC;
            empty : OUT STD_LOGIC
            );
        END component;
        
        
component STD_FIFO 
	Generic (
		constant DATA_WIDTH  : positive := 8;
		constant FIFO_DEPTH	: positive := 8
	);
	Port ( 
		CLK		: in  STD_LOGIC;
		RST		: in  STD_LOGIC;
		WriteEn	: in  STD_LOGIC;
		DataIn	: in  STD_LOGIC_VECTOR (DATA_WIDTH - 1 downto 0);
		ReadEn	: in  STD_LOGIC;
		DataOut	: out STD_LOGIC_VECTOR (DATA_WIDTH - 1 downto 0);
		Empty	: out STD_LOGIC;
		Full	: out STD_LOGIC
	);
end component;
         
        
constant clk_period: time:= 10 ns; 
signal   srst : std_logic; 
signal   clk  : std_logic;
signal   din  : std_logic_vector(7 downto 0); 
signal   wr_en: std_logic; 
signal   rd_en: std_logic;
signal   dout : std_logic_vector(7 downto 0); 
signal   full : std_logic;
signal   empty: std_logic; 

file file_VECTORS : text;
--file file_RESULTS : text;


begin

CLK_GEN:process
        begin 
        clk <= '1';
        wait for clk_period/2; 
        clk <= '0';
        wait for clk_period/2;
        end process; 

--FIFO_BLOCK: fifo_generator_0 port map (
--                                        clk   => clk,
--                                        rst   => srst,
--                                        din   => din,
--                                        wr_en => wr_en,
--                                        rd_en => rd_en,
--                                        dout  => dout,
--                                        full  => full,
--                                        empty => empty
--                                      );
                                      


FIFO_ENG: STD_FIFO generic map (DATA_WIDTH => 8, FIFO_DEPTH => 8)
                   	Port   map ( 
                                CLK		=> clk,
                                RST		=> srst,
                                WriteEn	=> wr_en,
                                DataIn	=> din,
                                ReadEn	=> rd_en,
                                DataOut	=> dout,
                                Empty	=> empty,
                                Full	=> full 
                            
                              );                                  
                                      
FSM:process
    variable v_iline : line; 
    variable v_oline: line; 
    variable v_din: std_logic_vector(7 downto 0); 
    begin 
    file_open(file_VECTORS, "/home/hadedy/Downloads/gage_hash_speed/GAGE_SPEED/GAGE_SPEED.srcs/sim_1/new/message.txt",  read_mode);
   -- file_open(file_RESULTS, "output_txt.txt", write_mode);
    srst <= '1';
    wait for clk_period; 
    srst <= '0';
    wr_en<= '0';
    rd_en<= '0';
    while not endfile(file_VECTORS) loop
      readline(file_VECTORS, v_Iline);
      hread(v_Iline, v_din);
      wr_en <= '1'; 
      din <= v_din; 
      wait for clk_period;  
       wr_en <= '0';

    end loop;
    file_close(file_VECTORS);
--    wait for clk_period; 
--    rd_en <= '1';
--    wait for clk_period; 
--    rd_en <= '1';
    wait for clk_period; 
    rd_en <= '0';

    wait for 15*clk_period; 
    rd_en <= '1';
    wait for clk_period; 
    rd_en <= '1';
    wait for 2*clk_period; 
    rd_en <= '1';
    wait for clk_period; 
    rd_en <= '0';
--    wait for clk_period; 
--    rd_en <= '1';
    wait for clk_period; 
    rd_en <= '0';

    
    wait;
  end process;
    

end Behavioral;
