----------------------------------------------------------------------------------
-- Project Name: Gage Hash Function, C: 224 and r: 8
-- 
-- Copyright 2019:
--     Mohamed El-Hadedy <mealy@cpp.edu>
--     Danilo Gligoroski <danilog@ntnu.no>
-- 
-- This program is free software; you can redistribute it and/or
-- modify it under the terms of the GNU General Public License as
-- published by the Free Software Foundation; either version 2 of the
-- License, or (at your option) any later version.
-- 
-- This program is distributed in the hope that it will be useful, but
-- WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
-- General Public License for more details.
-- 
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
-- 02110-1301, USA.
-- 
----------------------------------------------------------------------------------

----------------------------------------------------------------------------------
-- Company: California State Polytechnic University, Pomona, Norwegian University of Science and Technology (NTNU)
-- Engineer-1: Assistant Professor at ECE-department, College of Engineering 
-- Engineer-2: Professor at Department of Information Security and Communication Technology
-- 
-- Create Date: 06/22/2019 02:06:31 PM
-- Design Name-1: Mohamed El-Hadedy, also Airforce Research Lab(AFRL) fellow, US-AFA
-- Design Name-2: Danilo Gligoroski, also Visiting professor: Department of Mathematics and Statistics, University of South Florida, USA

-- Module Name: G_Engine - Behavioral
-- Project Name: Gage Hash Function , C: 224 and r: 8
-- Target Devices: xc7a35tcpg236-1 
-- Tool Versions: Vivado 2019.1
-- Description: Master Testbench
-- 
-- Dependencies: 
-- 
-- Revision:
-- Revision 0.01 - File Created
-- Additional Comments:
-- 
----------------------------------------------------------------------------------


library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use STD.textio.all;
use ieee.std_logic_textio.all;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.numeric_std.all;
use ieee.STD_LOGIC_ARITH.all;
use ieee.std_logic_unsigned.all;
use IEEE.math_real.all;

-- Uncomment the following library declaration if using
-- arithmetic functions with Signed or Unsigned values
--use IEEE.NUMERIC_STD.ALL;

-- Uncomment the following library declaration if instantiating
-- any Xilinx leaf cells in this code.
--library UNISIM;
--use UNISIM.VComponents.all;

entity gage1h256c224r008AllParallel_tb is
    generic (C: positive:= 224; r: positive:=8); 

--  Port ( );
end gage1h256c224r008AllParallel_tb;

architecture Behavioral of gage1h256c224r008AllParallel_tb is
component gage1h256c224r008AllParallel 
    generic (C: positive:= 224; r: positive:=8); 
    Port ( clk     : in  STD_LOGIC;
           rst     : in  STD_LOGIC;
           M_ST    : in  STD_LOGIC; 
           wr_en   : in  STD_LOGIC;
           Message : in  STD_LOGIC_VECTOR (7 downto 0);
           HASH_MSG: out STD_LOGIC_VECTOR (7 downto 0);
           HASH_OUT: out STD_LOGIC
         );
end component;

component STD_FIFO 
	Generic (
		constant DATA_WIDTH  : positive := 8;
		constant FIFO_DEPTH	: positive := 8
	);
	Port ( 
		CLK		: in  STD_LOGIC;
		RST		: in  STD_LOGIC;
		WriteEn	: in  STD_LOGIC;
		DataIn	: in  STD_LOGIC_VECTOR (DATA_WIDTH - 1 downto 0);
		ReadEn	: in  STD_LOGIC;
		DataOut	: out STD_LOGIC_VECTOR (DATA_WIDTH - 1 downto 0);
		Empty	: out STD_LOGIC;
		Full	: out STD_LOGIC
	);
end component;

constant clk_period: time:= 10 ns; 
signal   clk   : std_logic;
signal   rst   : std_logic; 
signal   M_ST  : std_logic;
signal   wr_en : std_logic; 
signal   MSG   : std_logic_vector(7 downto 0); 
signal HSH_MSG : std_logic_vector(7 downto 0); 
signal HASH_OUT: std_logic;
signal TMP     : std_logic_vector(255 downto 0);

file file_VECTORS : text;
file file_RESULTS : text;

signal count   : std_logic_vector(5 downto 0);


signal dout    :std_logic_vector(r-1 downto 0);

signal din     :std_logic_vector(r-1 downto 0); 
signal test_vector:std_logic_vector(511 downto 0);

signal HASH_FIN : std_logic; 
signal count_character: integer; 
signal incrment_msg:integer;
signal count_line     : integer;
signal full_tb, empty_tb: std_logic; 
signal wr_tb_en, rd_tb_en:std_logic; 
signal rst_tb:std_logic;

signal push_line:std_logic; 
signal init_count: std_logic_Vector(4 downto 0);
signal action_count:integer;

signal done_write:std_logic;
signal ext:std_logic;

type t_DATA is array (33 downto 0) of std_logic_vector(511 downto 0); 
type t_output_data is array(31 downto 0) of std_logic_vector(255 downto 0); 
type c_data is array (32 downto 0) of integer; 


signal array_testbench:t_DATA; 

signal output_array:t_output_data; 

signal test_wire:std_logic_vector(511 downto 0);

signal count_ch:c_data; 

signal tmp_output:std_logic_vector(255 downto 0); 
signal rst_output:std_logic;
 
signal tester:std_logic_vector(5 downto 0); 

begin

GENGINE_COMP: gage1h256c224r008AllParallel generic map( c=> c, r=>r)
                       port    map(
                                    clk      => clk,
                                    rst      => rst,
                                    M_ST     => M_ST,
                                    wr_en    => wr_en,
                                    Message  => MSG,
                                    HASH_MSG => HSH_MSG,
                                    HASH_OUT => HASH_OUT
                                  );
                                  
                                  
FIFO_ENG: STD_FIFO generic map (DATA_WIDTH => 8, FIFO_DEPTH => 64)
                   	Port   map ( 
                                CLK		=> clk,
                                RST		=> rst_tb,
                                WriteEn	=> wr_tb_en,
                                DataIn	=> din,
                                ReadEn	=> rd_tb_en,
                                DataOut	=> dout,
                                Empty	=> empty_tb,
                                Full	=> full_tb 
                            
                              );
                              
 
LOAD_DATA:process
         variable v_Iline: line; 
          variable V_SPACE: character; 
          variable v_din:std_logic_vector(1023 downto 0); 
          variable v_jump:integer;
          -- variable V_SPACE: character; 
          begin 
           count_character <= 0; 
           action_count    <= 0;
           rst_tb          <= '1';
           count_line <= 0;
           incrment_msg<= 0;
           wait for clk_period; 
           rst_tb <= '1';
           wait for clk_period; 
           rst_tb <= '0';
           file_open(file_VECTORS, "/home/hadedy/Downloads/gage1h256c224r008AllParallel/gage1h256c224r008AllParallel/gage1h256c224r008AllParallel.srcs/sim_1/new/messages.txt",  read_mode);
           while not endfile(file_VECTORS) and count_line <32 loop
         --  wait for clk_period; 
           readline(file_VECTORS, v_Iline);
           count_character <= v_Iline'length;
           wait for 1*clk_period;
           hread(v_Iline, v_din(511 downto (512-4*conv_integer(count_character))));
           -- wait for 1*clk_period;
           test_vector(511 downto (512-4*conv_integer(count_character))) <=v_din(511 downto (512-4*conv_integer(count_character)));
           array_testbench(conv_integer(count_line)) <=  test_vector;
           count_ch(conv_integer(count_line)) <= count_character;
           count_line <= count_line + 1;
           wait for 2*clk_period;
           end loop;
              -- wait for clk_period; 

           file_close(file_VECTORS); 
             M_ST <='0';
             wr_en<='0';
             rst  <='0';
             rst_output <= '0';
             rst_output <= '0';
             action_count <= 0;
          wait for clk_period;
          while count_line = 32 and action_count = 0 loop 
                    rst <= '1';
                    rst_output <= '1';
                    wait for clk_period;
                    rst <= '0';
                    rst_output <= '0';
                    wait for clk_period; 
                    M_ST <='1';
                    wait for 2*clk_period;
                    M_ST <='0';
                    while count <32 loop
                        wait for clk_period; 
                    end loop;
                    while count = 32 loop
                        action_count <= action_count +1;
                        wait for clk_period; 
 
                    end loop; 
            --    end loop;
         
              end loop; 
           wait for clk_period;
          while action_count >= 1 loop
          wait for 1000*clk_period; 
            rst  <= '1';
            wr_en<= '0';
            wait for 1*clk_period; 
            rst <= '0';
          --  wait for clk_period; 
          if count_ch(action_count)<61 then 
            for i in 1 to count_ch(action_count)/2 loop
                incrment_msg <= i; 
                wait for clk_period;
                wr_en <= '1';                   
                wait for clk_period;
                wr_en <= '0';
            end loop; 
                wr_en <= '0';            
              wait for clk_period; 
               M_ST <='1';
             wait for 2*clk_period;
               M_ST <='0';                  
                   while count <32 loop
                        wait for clk_period; 
                    end loop;
                    while count = 32 loop
                        action_count <= action_count +1;
                        wait for clk_period; 

                    end loop; 
            else 
                wr_en <= '0';
                M_ST  <= '0';
                wait; 
            end if; 
          end loop; 
                 
           

             
    end process;     
    
    
    
    LOADING_OUTPUT:process(clk,rst_tb)
                   begin 
                    if (rst_tb = '1') then
                        for j in 0 to 30 loop  
                        output_array(j)<=(others=>'0');
                        end loop; 
                        tester<=(others =>'0');
                    elsif (rising_edge(clk)) then 
                        if count_line = 32 and count =32 and action_count-1 < 31 then
                            output_array(action_count-1) <= tmp_output; 
                            tester <= tester + 1;
                        end if;
                    
                    end if;
                   end process; 
    
    
    
WORK_DATA:process
          variable v_oline: line; 
          variable V_SPACE: character; 
          variable v_jump:integer;
          variable count_end:integer; 
          begin 
               file_open(file_RESULTS, "/home/hadedy/Downloads/gage1h256c224r008AllParallel/gage1h256c224r008AllParallel/gage1h256c224r008AllParallel.srcs/sim_1/new/LWC_HASH_KAT_256.txt",  write_mode);
                   v_jump := 0;
                   count_end:= 0;
                    wait for clk_period; 
                    while (v_jump = 0 and count_line = 32) loop 
                    if tester = 31 then 
                         if count_end = 0 then 
                          write (v_oline,string'("Count = "));
                          write (v_oline, count_ch(count_end)+1); 
                          writeline(file_RESULTS, v_OLINE);
                          write (v_oline,string'("Msg = "));
                          write(v_oline, V_SPACE);
                          writeline(file_RESULTS, v_OLINE);
                          write (v_oline,string'("MD = "));
                          hwrite(v_oline,output_array(0),right);
                          writeline(file_RESULTS, v_OLINE);
                          writeline(file_RESULTS, v_OLINE);                    

                         v_jump := 0;
                          wait for clk_period; 
                         count_end := count_end + 1;

                        elsif(count_end >=1 and count_end <=31) then 
                         write (v_oline,string'("Count = "));
                          write (v_oline, (count_ch(count_end)/2)+1); 
                          writeline(file_RESULTS, v_OLINE);
                          wait for clk_period; 
                          write (v_oline,string'("Msg = "));
                          hwrite(v_oline, array_testbench(count_end+1)(511 downto (512-4*(count_ch(count_end)))));
                          writeline(file_RESULTS, v_OLINE);
                          wait for clk_period; 
                          write (v_oline,string'("MD = "));
                          hwrite(v_oline,output_array(count_end),right);
                          writeline(file_RESULTS, v_OLINE); 
                          writeline(file_RESULTS, v_OLINE);                    
                          wait for clk_period;  
                                
                       --     v_jump:=1; 
                       --     wait for clk_period;
                         count_end := count_end + 1;
                     
                        end if;

                    else 
                        wait for clk_period; 
                    end if; 

                    end loop;
                     file_close(file_RESULTS);                         

                    wait for clk_period; 

     
      end process;     
                                  
                                  
CLK_GEN:process
        begin 
        clk <= '1';
        wait for clk_period/2; 
        clk <= '0';
        wait for clk_period/2;
        end process; 
        
        
        
        
        
COUNT_GEN: process(clk,rst)
 
           begin 
 
           if  (rst = '1') then 
               count <= (others =>'0');
           elsif(rising_edge(clk)) then 
               if (HASH_OUT ='1')then 
                    count <= count +1;
               end if; 
                if count = 32 then 
                    count <= (others =>'0');
                end if; 
           end if;
           end process;    
           
WRITE_PROCESS: process
 
               begin 
 
                  --  wait for 1000*clk_period; 
             --     end if;  
                  
                  wait for clk_period;       
               end process;            
           
       
           
           process(clk,rst_output)
           begin 
            if (rst_output = '1') then 
                init_count <= (others =>'0');
                push_line  <= '0';
            elsif(rising_edge(clk)) then 
                if count = 32 then 
                --    push_line <='1';
                    init_count <= init_count + 1;
                end if; 
                if init_count > 1 then 
                  init_count <= init_count + 1;

                    push_line <= '1';
                else 
                    push_line <= '0';
                end if; 
               
                end if; 
           end process; 
           
DECO_TMP: process(count)
          begin
            case conv_integer(count) is
            
            when 1  => TMP(255 downto 248) <= HSH_MSG;
            when 2  => TMP(247 downto 240) <= HSH_MSG;
            when 3  => TMP(239 downto 232) <= HSH_MSG;
            when 4  => TMP(231 downto 224) <= HSH_MSG;
            when 5  => TMP(223 downto 216) <= HSH_MSG;
            when 6  => TMP(215 downto 208) <= HSH_MSG;
            when 7  => TMP(207 downto 200) <= HSH_MSG;
            when 8  => TMP(199 downto 192) <= HSH_MSG;
            when 9  => TMP(191 downto 184) <= HSH_MSG;
            when 10 => TMP(183 downto 176) <= HSH_MSG;
            when 11 => TMP(175 downto 168) <= HSH_MSG;
            when 12 => TMP(167 downto 160) <= HSH_MSG;
            when 13 => TMP(159 downto 152) <= HSH_MSG;
            when 14 => TMP(151 downto 144) <= HSH_MSG;
            when 15 => TMP(143 downto 136) <= HSH_MSG;
            when 16 => TMP(135 downto 128) <= HSH_MSG;
            when 17 => TMP(127 downto 120) <= HSH_MSG;
            when 18 => TMP(119 downto 112) <= HSH_MSG;
            when 19 => TMP(111 downto 104) <= HSH_MSG;
            when 20 => TMP(103 downto  96) <= HSH_MSG;
            when 21 => TMP(95  downto  88) <= HSH_MSG;
            when 22 => TMP(87  downto  80) <= HSH_MSG;
            when 23 => TMP(79  downto  72) <= HSH_MSG;
            when 24 => TMP(71  downto  64) <= HSH_MSG;
            when 25 => TMP(63  downto  56) <= HSH_MSG;
            when 26 => TMP(55  downto  48) <= HSH_MSG;
            when 27 => TMP(47  downto  40) <= HSH_MSG;
            when 28 => TMP(39  downto  32) <= HSH_MSG;
            when 29 => TMP(31  downto  24) <= HSH_MSG;
            when 30 => TMP(23  downto  16) <= HSH_MSG;
            when 31 => TMP(15  downto   8) <= HSH_MSG;
            when 32 => TMP(7   downto   0) <= HSH_MSG;
            
            when others => TMP <= (others =>'Z');
            end case;
          
          end process;                 
          
          
DECO_MSG: process(clk,rst, action_count, incrment_msg)

          begin
       --   if (rst = '1') then 
         --   MSG <= (others =>'0');
       --   elsif (rising_edge(clk))then 
            case conv_integer(incrment_msg) is
            when 1  => MSG<= array_testbench(conv_integer(action_count)+1)(511 downto 504) ; 
            when 2  => MSG<= array_testbench(conv_integer(action_count)+1)(503 downto 496) ; 
            when 3  => MSG<= array_testbench(conv_integer(action_count)+1)(495 downto 488) ; 
            when 4  => MSG<= array_testbench(conv_integer(action_count)+1)(487 downto 480) ; 
            when 5  => MSG<= array_testbench(conv_integer(action_count)+1)(479 downto 472) ; 
            when 6  => MSG<= array_testbench(conv_integer(action_count)+1)(471 downto 464) ; 
            when 7  => MSG<= array_testbench(conv_integer(action_count)+1)(463 downto 456) ; 
            when 8  => MSG<= array_testbench(conv_integer(action_count)+1)(455 downto 448) ; 
            when 9  => MSG<= array_testbench(conv_integer(action_count)+1)(447 downto 440) ; 
            when 10 => MSG<= array_testbench(conv_integer(action_count)+1)(439 downto 432) ;
            when 11 => MSG<= array_testbench(conv_integer(action_count)+1)(431 downto 424) ;
            when 12 => MSG<= array_testbench(conv_integer(action_count)+1)(423 downto 416) ;
            when 13 => MSG<= array_testbench(conv_integer(action_count)+1)(415 downto 408) ;
            when 14 => MSG<= array_testbench(conv_integer(action_count)+1)(407 downto 400) ;
            when 15 => MSG<= array_testbench(conv_integer(action_count)+1)(399 downto 392) ;
            when 16 => MSG<= array_testbench(conv_integer(action_count)+1)(391 downto 384) ;
            when 17 => MSG<= array_testbench(conv_integer(action_count)+1)(383 downto 376) ;
            when 18 => MSG<= array_testbench(conv_integer(action_count)+1)(375 downto 368) ;
            when 19 => MSG<= array_testbench(conv_integer(action_count)+1)(367 downto 360) ;
            when 20 => MSG<= array_testbench(conv_integer(action_count)+1)(359 downto 352) ;
            when 21 => MSG<= array_testbench(conv_integer(action_count)+1)(351 downto 344) ;
            when 22 => MSG<= array_testbench(conv_integer(action_count)+1)(343 downto 336) ;
            when 23 => MSG<= array_testbench(conv_integer(action_count)+1)(335 downto 328) ;
            when 24 => MSG<= array_testbench(conv_integer(action_count)+1)(327 downto 320) ;
            when 25 => MSG<= array_testbench(conv_integer(action_count)+1)(319 downto 312) ;
            when 26 => MSG<= array_testbench(conv_integer(action_count)+1)(311 downto 304) ;
            when 27 => MSG<= array_testbench(conv_integer(action_count)+1)(303 downto 296) ;
            when 28 => MSG<= array_testbench(conv_integer(action_count)+1)(295 downto 288) ;
            when 29 => MSG<= array_testbench(conv_integer(action_count)+1)(287 downto 280) ;
            when 30 => MSG<= array_testbench(conv_integer(action_count)+1)(279 downto 272) ;
            when 31 => MSG<= array_testbench(conv_integer(action_count)+1)(271 downto 264) ;
            when 32 => MSG<= array_testbench(conv_integer(action_count)+1)(263 downto 256) ;
            when 33 => MSG<= array_testbench(conv_integer(action_count)+1)(255 downto 248) ;
            when 34 => MSG<= array_testbench(conv_integer(action_count)+1)(247 downto 240) ;
            when 35 => MSG<= array_testbench(conv_integer(action_count)+1)(239 downto 232) ;
            when 36 => MSG<= array_testbench(conv_integer(action_count)+1)(231 downto 224) ;
            when 37 => MSG<= array_testbench(conv_integer(action_count)+1)(223 downto 216) ;
            when 38 => MSG<= array_testbench(conv_integer(action_count)+1)(215 downto 208) ;
            when 39 => MSG<= array_testbench(conv_integer(action_count)+1)(207 downto 200) ;
            when 40 => MSG<= array_testbench(conv_integer(action_count)+1)(199 downto 192) ;
            when 41 => MSG<= array_testbench(conv_integer(action_count)+1)(191 downto 184) ;
            when 42 => MSG<= array_testbench(conv_integer(action_count)+1)(183 downto 176) ;
            when 43 => MSG<= array_testbench(conv_integer(action_count)+1)(175 downto 168) ;
            when 44 => MSG<= array_testbench(conv_integer(action_count)+1)(167 downto 160) ;
            when 45 => MSG<= array_testbench(conv_integer(action_count)+1)(159 downto 152) ;
            when 46 => MSG<= array_testbench(conv_integer(action_count)+1)(151 downto 144) ;
            when 47 => MSG<= array_testbench(conv_integer(action_count)+1)(143 downto 136) ;
            when 48 => MSG<= array_testbench(conv_integer(action_count)+1)(135 downto 128) ;
            when 49 => MSG<= array_testbench(conv_integer(action_count)+1)(127 downto 120) ;
            when 50 => MSG<= array_testbench(conv_integer(action_count)+1)(119 downto 112) ;
            when 51 => MSG<= array_testbench(conv_integer(action_count)+1)(111 downto 104) ;
            when 52 => MSG<= array_testbench(conv_integer(action_count)+1)(103 downto  96) ;
            when 53 => MSG<= array_testbench(conv_integer(action_count)+1)(95  downto  88) ;
            when 54 => MSG<= array_testbench(conv_integer(action_count)+1)(87  downto  80) ;
            when 55 => MSG<= array_testbench(conv_integer(action_count)+1)(79  downto  72) ;
            when 56 => MSG<= array_testbench(conv_integer(action_count)+1)(71  downto  64) ;
            when 57 => MSG<= array_testbench(conv_integer(action_count)+1)(63  downto  56) ;
            when 58 => MSG<= array_testbench(conv_integer(action_count)+1)(55  downto  48) ;
            when 59 => MSG<= array_testbench(conv_integer(action_count)+1)(47  downto  40) ;
            when 60 => MSG<= array_testbench(conv_integer(action_count)+1)(39  downto  32) ;
            when 61 => MSG<= array_testbench(conv_integer(action_count)+1)(31  downto  24) ;
            when 62 => MSG<= array_testbench(conv_integer(action_count)+1)(23  downto  16) ;
            when 63 => MSG<= array_testbench(conv_integer(action_count)+1)(15  downto   8) ;
            when 64 => MSG<= array_testbench(conv_integer(action_count)+1)(7   downto   0) ;
            when others => MSG <= (others =>'Z');
            end case;
     --     end if;
          end process;                 
                                 
        
OUTPUT:process(clk,rst_output) 
       begin 
        if(rst_output = '1') then 
            tmp_output <=(others =>'0');
        else 
          if count = 32 then 
            tmp_output <= tmp; 
          end if;         
        
        end if; 
       
       end process; 




 
 
    

end Behavioral;

